import 'babel-polyfill';

import path          from 'path';
import fs            from 'fs';
import { spawn }     from 'child_process';
import portscanner   from 'portscanner';
import ip            from 'ip';

const VORLON_PATH        = path.join(__dirname, '../vorlon');
const VORLON_BIN         = path.join(VORLON_PATH, '/bin/vorlon');
const CONFIGURE_FILE     = path.join(VORLON_PATH, '/Server/config.json');
const CONFIGURE_BAK_FILE = `${CONFIGURE_FILE}.bak`;
const VORLON_PARAMS      = [
  'port',
  'proxyPort',
  'state',
];

export default class Vorlon {
  setting = {
    runServer       : true,
    appendScriptTag : true,
    port            : undefined,
    proxyPort       : undefined,
    state           : '',
  };

  _defaults (options = {}) {
    return Object.assign({}, this.setting, options);
  }

  constructor (options = {}) {
    this.setting = this._defaults(options);
  }

  apply (compiler) {
    let options = this._defaults();

    this.compiler = compiler;

    options.runServer && compiler.plugin('watch-run', this.start.bind(this));
    compiler.plugin('compilation', this.applyCompilation.bind(this));
  }

  applyCompilation (compilation) {
    compilation.mainTemplate.plugin('startup', this.scriptTag.bind(this));
  }

  async start (watching, done) {
    const defaults      = this.setting;
    const idlePort      = defaults.idlePort || await getIdlePort();
    const idleProxyPort = defaults.idleProxyPort || await getIdlePort(idlePort +1);

    let options = this._defaults({
      port      : idlePort,
      proxyPort : idleProxyPort,
    });

    /**
     * Vorlon 没有配置项
     * 因此必须修改 /path/to/vorlon/Server/config.json 里面
     * 进行配置项修改
     *
     * 这里会保存原先的文件到 .bak 里面, 写入新的配置到配置文件中
     */

    if (!fs.existsSync(CONFIGURE_BAK_FILE)) {
      fs.copySync(CONFIGURE_FILE, CONFIGURE_BAK_FILE);
    }

    let buffer = fs.readFileSync(CONFIGURE_FILE);
    let source = JSON.parse(buffer);

    Object.assign(source, options);
    fs.writeFileSync(CONFIGURE_FILE, JSON.stringify(source));

    /**
     * 启动 vorlon
     */

    let params = VORLON_PARAMS
    .map(function (key) {
      return `--${key} ${options[key]}`;
    })
    .join(' ')
    .split(' ');

    let childProgress = spawn(VORLON_BIN, params, {
      stdio: 'inherit',
    });

    childProgress.on('close', () => {
      this.reset();
    });

    this.server        = childProgress;
    this.serverOptions = options;

    done();
  }

  scriptTag (source) {
    let options = this.serverOptions || this.setting || {};

    if (options.appendScriptTag) {
      let injector = this.autoloadScript();
      return injector + source;
    }

    return source;
  }

  autoloadScript () {
    const options  = this.serverOptions || this.setting || {};
    const servPort = options.port;

    if ('number' !== typeof servPort) {
      return logger('error')('No script tag will inserted, because the config is not run vorlon server `(runServer)`, and you also do not configure `port` and `proxyPort`.');
    }

    const servIp = ip.address();
    const client = `http://${servIp}:${servPort}/vorlon.js`;
    const server = `http://${servIp}:${servPort}/dashboard`;

    return `
      // Vorlon plugin
      !(function () {
        if ('undefined' === typeof window) {
          return;
        }

        var id = 'webpack-vorlon-plugin-script';
        var vorlon = document.createElement('script');
        vorlon.id  = id;
        vorlon.src = '${client}';
        document.head.appendChild(vorlon);
      })();

      ${logger()('`Vorlon Server`: ' + server)}
    `;

    function logger (type = 'log') {
      return function (...args) {
        let colors = [];
        let params = args.map(function (param) {
          while (/`([\w\W]+?)`/.exec(param)) {
            param = param.replace(/`([\w\W]+?)`/, '%c$1%c');
            colors.push('"color:#0073b7;font-weight:bolder;"', '""');
          }

          return `"${param}"`
        });

        params = params.concat(colors);

        return `
          !(function () {
            if ('undefined' === typeof window) {
              return;
            }

            var log = document.createElement('script');
            log.innerHTML = '"function" === typeof console.${type} && console.${type}(${params})';
            document.head.appendChild(log);
          })();
        `;

        return '';
      };
    }
  }

  reset () {
    if (this.server) {
      this.server.kill();

      if (fs.existsSync(CONFIGURE_BAK_FILE)) {
        fs.copySync(CONFIGURE_FILE, CONFIGURE_BAK_FILE);
      }

      this.server        = undefined;
      this.serverOptions = undefined;
    }
  }
}

function getIdlePort (port = 50000) {
  return new Promise((resolve, reject) => {
    portscanner.findAPortNotInUse(port, port + 10000, '0.0.0.0', (error, port) => {
      error ? reject(error) : resolve(port);
    });
  });
}
